/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef DATAFRAME_ELEMENT_PROXY_HPP
#define DATAFRAME_ELEMENT_PROXY_HPP 1

#include <type_traits>
#include <boost/operators.hpp>
#include <dataframe/element_transform.hpp>
#include <observer_ptr.hpp>

namespace dataframe
{

  template < typename T
	     , typename U = T
	     >
  class element_proxy
    : boost::totally_ordered< element_proxy<T,U>
    , boost::totally_ordered< element_proxy<T,U>,T
    > >
  {
    using self = element_proxy;
    using pointer = observer_ptr<T>;

  public:
    element_proxy () = default;
    
    explicit element_proxy (T& t)
      : p_(make_observer(t))
    {}

    self& operator = (T const& t)
    {
      *p_ = t;
      return *this;
    }

    self& operator = (T&& t)
    {
      *p_ = std::move(t);
      return *this;
    }

    auto value () const
    {
      using transform = element_transform<T,U>;
      return transform::apply(*p_);
    }

    operator T& () const { return *p_; }

  private:
    pointer p_;
  };

  template < typename T, typename U >
  class element_proxy< T const,U >
  {
    using pointer = observer_ptr<T const>;

  public:
    element_proxy () = default;
    
    explicit element_proxy (T const& t)
      : p_(make_observer(t))
    {}

    auto value () const
    {
      using transform = element_transform<T const,U>;
      return transform::apply(*p_);
    }

    operator T const& () const { return *p_; }

  private:
    pointer p_;
  };

  template < typename T, typename U >
  bool operator == (element_proxy<T,U> left, element_proxy<T,U> right)
  { return left.value() == right.value(); }

  template < typename T, typename U >
  bool operator < (element_proxy<T,U> left, element_proxy<T,U> right)
  { return left.value() < right.value(); }

  template < typename T, typename U >
  bool operator == (element_proxy<T,U> left, T right)
  { return static_cast<T>(left) == right; }

  template < typename T, typename U >
  bool operator < (element_proxy<T,U> left, T right)
  { return static_cast<T>(left) < right; }

  template < typename T, typename U >
  bool operator > (element_proxy<T,U> left, T right)
  { return static_cast<T>(left) > right; }

  template < typename T, typename U, typename V >
  std::enable_if_t<!std::is_same<T,U>::value,bool>
  operator == (element_proxy<T,U> left, V right)
  { return left.value() == right; }

  template < typename T, typename U, typename V >
  std::enable_if_t<!std::is_same<T,U>::value,bool>
  operator != (element_proxy<T,U> left, V right)
  { return left.value() == right; }

  template < typename T, typename U, typename V >
  std::enable_if_t<!std::is_same<T,U>::value,bool>
  operator < (element_proxy<T,U> left, V right)
  { return left.value() < right; }

  template < typename T, typename U, typename V >
  std::enable_if_t<!std::is_same<T,U>::value,bool>
  operator <= (element_proxy<T,U> left, V right)
  { return left.value() <= right; }

  template < typename T, typename U, typename V >
  std::enable_if_t<!std::is_same<T,U>::value,bool>
  operator >= (element_proxy<T,U> left, V right)
  { return left.value() >= right; }

  template < typename T, typename U, typename V >
  std::enable_if_t<!std::is_same<T,U>::value,bool>
  operator > (element_proxy<T,U> left, V right)
  { return left.value() > right; }

  template < typename T, typename U >
  std::ostream& operator << (std::ostream& os, element_proxy<T,U> proxy)
  { return os << proxy.value(); }

  template < typename U, typename T >
  auto make_element_proxy (T&& t)
  {
    using element_proxy_type = element_proxy<std::remove_reference_t<T>,U>;
    return element_proxy_type(std::forward<T>(t));
  }

} // namespace dataframe

#endif
