/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_DISTRIBUTION_MCMC_METROPOLIS_SAMPLER_HPP
#define GRAPH_MODEL_DISTRIBUTION_MCMC_METROPOLIS_SAMPLER_HPP 1

#include <algorithm>
#include <functional>
#include <vector>
#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>
#include <boost/thread/thread.hpp>
#include <graph_model/MCMC/metropolis_sampler_state.hpp>
#include <graph_model/MCMC/random_variable_proxy.hpp>
#include <graph_model/MCMC/random_variable_proxy_collection.hpp>
#include <graph_model/concept/compatible_rv_sampler_concept.hpp>
#include <graph_model/random/uniform_01.hpp>
#include <graph_model/signals/default_signal_policy.hpp>
#include <graph_model/vertex/expression_function.hpp>
#include <logging/channel_log.hpp>
#include <logging/global_log.hpp>
#include <signals/tags.hpp>

namespace graph_model
{

  namespace MCMC
  {

    /// Metropolis sampler policy
    /// \tparam Expr The type of expression to evaluate
    /// \tparam SignalPolicy The policy controlling emission of signals
    ///
    /// A Metropolis sampler assumes that proposals are generated from
    /// a symmetric distribution, i.e., one for which
    /// Pr(proposal|current) == Pr(current|proposal).  Consequently,
    /// acceptance of the proposal depends only on the probabilities
    /// of the current and proposed states.  These probabilities are
    /// obtained by evaluating the expression used to construct the
    /// sampler, before and after updating the proposed variate for
    /// each sampled random variable.  The sampler may update any
    /// number of random variables, each one of which must be added to
    /// the sampler (along with the expression from which proposals
    /// are drawn) via the \c add_rv() member function.  Proposals
    /// for all such random variables will be either accepted or
    /// rejected together based upon the ratio of the expression
    /// evaluation before and after the proposals are generated.
    ///
    template < typename Expr
	       , typename SignalPolicy = signals::default_signal_policy
	       >
    class metropolis_sampler_policy
    {
      using rv_proxy_type = random_variable_proxy;
      using factory_type = std::function<rv_proxy_type()>;
      using factory_vector_type = std::vector<factory_type>;
      using log_type = logging::channel_log;

    public:
      metropolis_sampler_policy () = default;

      /// Construct a metropolis_sampler_policy
      /// \tparam E The type of expression to evaluate for the probability
      ///
      /// Construct a metropolis_sampler_policy from an expression for
      /// the probability.  The expression is interpreted as the
      /// probability (or likelihood) of the current and proposed
      /// variates.  For simple use cases, i.e., sampling a single
      /// random variable, the expression will normally be the Markov
      /// blanket of the random variable.
      ///
      template < typename E >
      metropolis_sampler_policy (E&& expr)
	: expr_(std::forward<E>(expr))
      {}

      template < typename E >
      metropolis_sampler_policy (E& expr, SignalPolicy&& signal_policy)
	: expr_(std::forward<E>(expr))
	, signal_policy_(std::move(signal_policy))
      {}

      /// \brief Add a new random variable to the sampler
      /// \tparam RV The type of random variable to sample
      /// \tparam E An expression that generates proposals
      ///
      template < typename RV, typename E >
      auto& add_rv (RV& rv, E&& expr) &
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "add_rv<RV,E>(RV&,E&&)& [" << this << "]: rv.identifier=" << rv.identifier();
	add_factory(rv,std::forward<E>(expr));
	return *this;
      }

      /// \brief Add a new random variable to the sampler
      /// \tparam RV The type of random variable to sample
      /// \tparam E An expression that generates proposals
      ///
      template < typename RV, typename E >
      auto&& add_rv (RV& rv, E&& expr) &&
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "add_rv<RV,E>(RV&,E&&)&& [" << this << "]: rv.identifier=" << rv.identifier();
	add_factory(rv,std::forward<E>(expr));
	return std::move(*this);
      }

      void operator () () { initialize_or_update(); }

      template < typename Engine >
      void operator () (Engine& engine) { initialize_or_update(engine); }

      void initialize_or_update ()
      {
	if (initialized_)
	  update();
	else
	  {
	    initialize();
	    initialized_ = true;
	  }
      }

      template < typename Engine >
      void initialize_or_update (Engine& engine)
      {
	if (initialized_)
	  update(engine);
	else
	  {
	    initialize();
	    initialized_ = true;
	  }
      }

      // XXX - this should not be here? initialize elsewhere or provide a functor
      void initialize ()
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "initialize() [" << this << "]";
	auto proxies = make_proxies();
	proxies.use_proposed_variate();
	long n = 0;
	while (markov_blanket() == 0)
	  {
	    proxies.clear();
	    proxies = make_proxies();
	    ++n;
	    assert(n <= 1000);
	  }
	proxies.accept();
	proxies.use_current_variate();
	emit<signals::tag::iteration>();
	emit<signals::tag::accept>();
	GLOBAL_LOG(log_,trace) << "initialize() [" << this << "] completed";
      }

      void update ()
      { update_(make_metropolis_sampler_state(markov_blanket(),random::uniform_01<>::apply())); }

      template < typename Engine >
      void update (Engine& engine)
      { update_(make_metropolis_sampler_state(markov_blanket(),random::uniform_01<>::apply(engine))); }

      void commit ()
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "commit() [" << this << "]";
	make_proxies().commit();
      }

      void graph_callback (vertex_and_edge_list_graph& graph)
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "graph_callback() [" << this << "]: &graph=" << &graph;
	make_proxies().graph_callback(graph);
	GLOBAL_LOG(log_,trace) << "graph_callback() [" << this << "]: &graph=" << &graph << ", completed";
      }

      template < typename SignalTag, typename Slot >
      auto connect (Slot&& slot)
      { return signal_policy_.template connect<SignalTag>(std::forward<Slot>(slot)); }

      template < typename SignalTag, typename Slot >
      auto&& connect_signal (Slot&& slot) &&
      { connect<SignalTag>(std::forward<Slot>(slot)); return std::move(*this); }

    private:
      template < typename RV, typename E >
      void add_factory (RV& rv, E&& expr)
      {
	using vertex::make_expression_function;

	auto expression_function = make_expression_function(std::forward<E>(expr));
	BOOST_CONCEPT_ASSERT((concept::compatible_rv_sampler_concept<RV,decltype(expression_function)>));

	factories_.push_back
	  ([&rv,f=expression_function]()
	   {
	     return make_random_variable_proxy
	       (rv,
		[f](auto& rv)
		{
		  rv.reset_proposed_variate(f());
		});
	   });
      }

      auto make_proxies ()
      {
	using std::for_each;
	random_variable_proxy_collection proxies;
	for_each(factories_.begin(),factories_.end(),
		 [&proxies](auto& factory){ proxies.push_back(factory()); });
	return proxies;
      }

      template < typename SamplerState >
      void update_ (SamplerState state)
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "update_<SamplerState>(SamplerState) [" << this << "]";
	auto proxies = make_proxies();
	proxies.use_proposed_variate();
	state.proposal_pdf(markov_blanket());
	GLOBAL_LOG(log_,trace) << state;
	if (state.accept())
	  {
	    proxies.accept();
	    emit<signals::tag::accept>();
	  }
	else
	  {
	    proxies.reject();
	    emit<signals::tag::reject>();
	  }
	proxies.use_current_variate();
	emit<signals::tag::iteration>();
	GLOBAL_LOG(log_,trace) << "update_<SamplerState>(SamplerState) [" << this << "] completed";
      }

    private:
      template < typename SignalTag >
      void emit () const { signal_policy_.template emit<SignalTag>(); }

      template < typename ... T >
      auto markov_blanket (T&& ... t) const { return expr_(std::forward<T>(t)...); }

    private:
      bool initialized_ = false;
      Expr expr_;
      SignalPolicy signal_policy_;
      factory_vector_type factories_;
      mutable log_type log_ = "graph_model::MCMC::metropolis_sampler_policy";
    };

    template < typename Expr >
    auto
    make_metropolis_sampler_policy (Expr&& expr)
    {
      using vertex::make_expression_function;
      auto expression = make_expression_function(std::forward<Expr>(expr));
      using policy_type = metropolis_sampler_policy<decltype(expression)>;
      return policy_type(expression);
    }

    template < typename Expr, typename SignalPolicy >
    auto
    make_metropolis_sampler_policy (Expr&& expr, SignalPolicy&& signal_policy)
    {
      using vertex::make_expression_function;
      auto expression = make_expression_function(std::forward<Expr>(expr));
      using policy_type = metropolis_sampler_policy<decltype(expression),std::decay_t<SignalPolicy>>;
      return policy_type(expression,std::forward<SignalPolicy>(signal_policy));
    }

    template < typename SignalPolicy, typename Expr >
    auto
    make_metropolis_sampler_policy (Expr&& expr)
    {
      using vertex::make_expression_function;
      auto expression = make_expression_function(std::forward<Expr>(expr));
      using policy_type = metropolis_sampler_policy<decltype(expression),SignalPolicy>;
      return policy_type(expression);
    }

    template < typename Expr >
    auto metropolis_sampler (Expr&& expr)
    {
      return make_metropolis_sampler_policy(std::forward<Expr>(expr));
    }

    template < typename Expr, typename SignalPolicy >
    auto metropolis_sampler (Expr&& expr, SignalPolicy&& signal_policy)
    {
      return make_metropolis_sampler_policy(std::forward<Expr>(expr),std::forward<SignalPolicy>(signal_policy));
    }

    template < typename SignalPolicy, typename Expr >
    auto metropolis_sampler (Expr&& expr)
    {
      return make_metropolis_sampler_policy<SignalPolicy>(std::forward<Expr>(expr));
    }

  } // namespace MCMC

} // namespace graph_model

#endif
