/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_DISTRIBUTION_MULTINOMIAL_MULTINOMIAL_STRATEGY_HPP
#define GRAPH_MODEL_DISTRIBUTION_MULTINOMIAL_MULTINOMIAL_STRATEGY_HPP 1

#include <type_traits>
#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>

#include <graph_model/distribution/distribution_policy.hpp>
#include <graph_model/random/distribution.hpp>
#include <graph_model/vertex/parameter_tags.hpp>
#include <graph_model/vertex/tagged_parameter.hpp>
#include <sample_space/is_possible.hpp>

#include <graph_model/distribution/parameter/parameter_pack.hpp>
#include <graph_model/distribution/parameter/probability.hpp>
#include <graph_model/distribution/parameter/variate.hpp>

#include <sample_space/math/multinomial_distribution.hpp>
#include <sample_space/random/multinomial_distribution.hpp>

#include <traits/probability.hpp>

#include <logging/channel_log.hpp>
#include <logging/global_log.hpp>

namespace graph_model
{

  namespace distribution
  {

    namespace multinomial
    {

      /// \brief Generate a multinomial distribution
      ///
      /// This policy class models the Distribution Generator concept,
      /// which generates a distribution from its parameters.  A
      /// distribution models the corresponding concept in the Boost
      /// Math library.  Specifically, it can be evaluated by the \c
      /// pdf() function (and possibly the \c cdf() function if it is
      /// a scalar distribution).
      ///
      /// This policy generates a multinomial distribution.
      ///
      struct distribution_policy
      {
	template < typename Variate, typename P >
	static auto apply (P p)
	{
	  using probability_type = ::traits::probability_t<P>;
	  using multinomial_distribution = sample_space::math::multinomial_distribution<probability_type,P>;
	  return multinomial_distribution(p);
	}
      };

      /// \brief Generate a multinomial random number generator
      ///
      /// This policy class generates a multinomial distribution that
      /// models the Random Distribution concept from the Boost Random
      /// library.
      ///
      struct generator_policy
      {
	template < typename P >
	static auto& apply ()
	{
	  using multinomial_distribution_type = sample_space::random::multinomial_distribution<P>;
	  static multinomial_distribution_type multinomial_distribution;
	  return multinomial_distribution;
	}
      };

      /// \brief Strategy for multinomial distributions
      /// \tparam ParameterPack The parameters
      ///
      template < typename ParameterPack >
      class multinomial_strategy
      {
	using policy_type = distribution::distribution_policy<distribution_policy,generator_policy>;
	using log_type = logging::channel_log;

      public:
	multinomial_strategy () = default;
	template < typename Tags, typename Defaults, typename Tagged, typename Untagged >
	explicit multinomial_strategy (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack)
	  : pack_(pack)
	{}

      public:
	template < typename ... T >
	auto variate (T&& ... t) const
	{ return pack_.template value<parameter::tag::variate>(std::forward<T>(t)...); }

	template < typename ... T >
	auto probability (T&& ... t) const
	{ return pack_.template value<parameter::tag::probability>(std::forward<T>(t)...); }

      public:
	template < typename ... T >
	auto sample (T&& ... t) const
	{
	  using namespace boost::log::trivial;

	  auto p = probability(std::forward<T>(t)...);
	  auto result = sample(p);

	  GLOBAL_LOG(log_,trace)
	    << "p=" << p << ", "
	    << "sample=" << result;

	  return result;
	}
	template < typename ... T >
	auto pdf (T&& ... t) const
	{
	  using namespace boost::log::trivial;
	  
	  auto variate = this->variate(std::forward<T>(t)...);
	  auto p = probability(std::forward<T>(t)...);
	  auto result = pdf(variate,p);

	  GLOBAL_LOG(log_,trace)
	    << "variate=" << variate << ", "
	    << "p=" << p << ", "
	    << "pdf=" << result;

	  return result;
	}

      private:
	template < typename P >
	auto sample (P p) const
	{
	  using sample_space::is_possible;
	  assert(is_possible(p));
	  return policy_type::sample(p);
	}

	template < typename Variate, typename P >
	auto pdf (Variate variate, P p) const
	{
	  using namespace boost::log::trivial;
	  using sample_space::is_possible;

	  auto result =
	    is_possible(variate)
	    && is_possible(p)
	    ? policy_type::eval_pdf(variate,p)
	    : 0;

	  GLOBAL_LOG(log_,trace)
	    << "variate=" << variate << ", "
	    << "p=" << p << ", "
	    << "pdf=" << result;

	  return result;
	}

      private:
	ParameterPack pack_;
	mutable log_type log_ = "graph_model::distribution::multinomial";
      };

      template < typename ParameterPack >
      auto make_multinomial_strategy (ParameterPack parameter_pack)
      {
	using multinomial_strategy_type = multinomial_strategy<ParameterPack>;
	return multinomial_strategy_type(parameter_pack);
      };

    } // namespace multinomial

  } // namespace distribution

} // namespace graph_model

#endif
