/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file
/// \brief Construct a strategy for a normal distribution
///

#ifndef GRAPH_MODEL_DISTRIBUTION_NORMAL_NORMAL_DISTRIBUTION_HPP
#define GRAPH_MODEL_DISTRIBUTION_NORMAL_NORMAL_DISTRIBUTION_HPP 1

#include <boost/hana/map.hpp>
#include <graph_model/distribution/normal/normal_strategy.hpp>
#include <graph_model/distribution/parameter/dispersion.hpp>
#include <graph_model/distribution/parameter/make_parameter_default.hpp>
#include <graph_model/distribution/parameter/mean.hpp>
#include <graph_model/distribution/parameter/named_parameter.hpp>
#include <graph_model/distribution/parameter/parameter_pack.hpp>
#include <graph_model/distribution/parameter/precision.hpp>
#include <graph_model/distribution/parameter/standard_deviation.hpp>
#include <graph_model/distribution/parameter/variance.hpp>
#include <graph_model/distribution/parameter/variate.hpp>

namespace graph_model
{

  namespace distribution
  {

    namespace normal
    {

      namespace detail
      {

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   >
	auto normal_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>);

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Key, typename Value
		   , typename ... Parameters
		   >
	auto normal_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>,
	 parameter::named_parameter<Key,Value>,
	 Parameters&& ...);

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Value
		   , typename ... Parameters
		   >
	auto normal_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>,
	 Value&&,
	 Parameters&& ...);

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged >
	auto normal_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>);

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Key, typename Value
		   , typename ... Parameters
		   >
	auto normal_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>,
	 parameter::named_parameter<Key,Value>,
	 Parameters&& ...);

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Value
		   , typename ... Parameters
		   >
	auto normal_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>,
	 Value&&,
	 Parameters&& ...);

	template < typename Tag >
	struct retag
	{
	  template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		     , typename Value
		     , typename ... Parameters
		   >
	  static auto apply
	  (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	   parameter::named_parameter<Tag,Value> parameter,
	   Parameters&& ... parameters)
	  {
	    return normal_distribution
	      (pack.add(parameter),
	       std::forward<Parameters>(parameters)...);
	  }
	};

	template <>
	struct retag < parameter::tag::standard_deviation >
	{
	  template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		     , typename Value
		     , typename ... Parameters
		   >
	  static auto apply
	  (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	   parameter::named_parameter<parameter::tag::standard_deviation,Value> parameter,
	   Parameters&& ... parameters)
	  {
	    return normal_distribution<parameter::tag::standard_deviation>
	      (pack.add(parameter::retag<parameter::tag::dispersion>(parameter)),
	       std::forward<Parameters>(parameters)...);
	  }
	};

	template <>
	struct retag < parameter::tag::variance >
	{
	  template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		     , typename Value
		     , typename ... Parameters
		   >
	  static auto apply
	  (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	   parameter::named_parameter<parameter::tag::variance,Value> parameter,
	   Parameters&& ... parameters)
	  {
	    return normal_distribution<parameter::tag::variance>
	      (pack.add(parameter::retag<parameter::tag::dispersion>(parameter)),
	       std::forward<Parameters>(parameters)...);
	  }
	};

	template <>
	struct retag < parameter::tag::precision >
	{
	  template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		     , typename Value
		     , typename ... Parameters
		   >
	  static auto apply
	  (parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	   parameter::named_parameter<parameter::tag::precision,Value> parameter,
	   Parameters&& ... parameters)
	  {
	    return normal_distribution<parameter::tag::precision>
	      (pack.add(parameter::retag<parameter::tag::dispersion>(parameter)),
	       std::forward<Parameters>(parameters)...);
	  }
	};

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   >
	auto normal_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack)
	{ return make_normal_strategy<Tag>(pack); }

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Key, typename Value
		   , typename ... Parameters
		   >
	auto normal_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	 parameter::named_parameter<Key,Value> parameter,
	 Parameters&& ... parameters)
	{
	  return normal_distribution<Tag>
	    (pack.add(parameter),
	     std::forward<Parameters>(parameters)...);
	}

	template < typename Tag
		   , typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Value
		   , typename ... Parameters
		   >
	auto normal_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	 Value&& value,
	 Parameters&& ... parameters)
	{
	  return normal_distribution<Tag>
	    (pack.add(std::forward<Value>(value)),
	     std::forward<Parameters>(parameters)...);
	}

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged >
	auto normal_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack)
	{ return normal_distribution<parameter::tag::standard_deviation>(pack); }

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Key, typename Value
		   , typename ... Parameters
		   >
	auto normal_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	 parameter::named_parameter<Key,Value> parameter,
	 Parameters&& ... parameters)
	{
	  return detail::retag<Key>::apply(pack,parameter,std::forward<Parameters>(parameters)...);
	}

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Value
		   , typename ... Parameters
		   >
	auto normal_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	 Value&& value,
	 Parameters&& ... parameters)
	{
	  return normal_distribution
	    (pack.add(std::forward<Value>(value)),
	     std::forward<Parameters>(parameters)...);
	}

      }	// namespace detail

      template < typename ... Parameters >
      auto normal_distribution (Parameters&& ... parameters)
      {
	using boost::hana::make_map;
	using parameter::make_parameter;
	using parameter::make_parameter_pack;
	using vertex::make_expression_function;
      	return detail::normal_distribution
	  (make_parameter_pack
	   (make_parameter<parameter::tag::mean>().default_value(0),
	    make_parameter<parameter::tag::dispersion>().default_value(1),
	    make_parameter<parameter::tag::variate>().default_value(0)),
	   std::forward<Parameters>(parameters)...);
      }

    } // namespace normal

  } // namespace distribution

} // namespace graph_model

#endif
