/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_GRAPH_ADD_CONNECTED_COMPONENTS_BASE_HPP
#define GRAPH_MODEL_GRAPH_ADD_CONNECTED_COMPONENTS_BASE_HPP 1

#include <map>
#include <boost/graph/breadth_first_search.hpp>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/properties.hpp>
#include <graph_model/graph/color_map.hpp>
#include <graph_model/traits/edge_descriptor.hpp>
#include <graph_model/graph/graph_base.hpp>
#include <graph_model/traits/vertex_descriptor.hpp>

namespace graph_model
{

  namespace graph
  {

    template < typename Derived, typename Tag, typename Base = graph_base >
    class add_connected_components_base
      : public Base
    {
      using vertex_descriptor = traits::vertex_descriptor_t<Tag>;
      using edge_descriptor = traits::edge_descriptor_t<Tag>;

    public:
      add_connected_components_base () = default;
      template < typename Graph >
      add_connected_components_base (Graph const& g, vertex_descriptor source)
	: Base(g,source)
      {}

      template < typename Graph >
      auto& add_connected_component (Graph const& g,
				     typename boost::graph_traits<Graph>::vertex_descriptor s)
      {
	static_assert(std::is_same<typename boost::graph_traits<Graph>::vertex_descriptor,vertex_descriptor>::value,
		      "Mismatched vertex descriptors");

	using color_type = boost::default_color_type;
	using color_map_type = std::map<vertex_descriptor,color_type>;
      
	using std::make_pair;
	using boost::breadth_first_visit;

	color_map_type color_map;
	breadth_first_visit
	  (g,s,
	   boost::visitor(derived().bfs_visitor())
	   .color_map(make_color_map(color_map)));
	return derived();
      }

      template < typename Graph >
      auto& add_connected_components (Graph const& g) { return derived(); }

      template < typename Graph
		 , typename ... VertexDescriptors
		 >
      auto& add_connected_components (Graph const& g, typename boost::graph_traits<Graph>::vertex_descriptor source, VertexDescriptors&& ... others)
      {
	add_connected_component(g,source);
	return add_connected_components(g,std::forward<VertexDescriptors>(others)...);
      }

    private:
      Derived& derived () { return static_cast<Derived&>(*this); }
      Derived const& derived () const { return static_cast<Derived const&>(*this); }
    };

  } // namespace graph

} // namespace graph_model

#endif
