/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_GRAPH_GRAPHVIZ_HPP
#define GRAPH_MODEL_GRAPH_GRAPHVIZ_HPP 1

#include <iosfwd>
#include <string>
#include <boost/filesystem/fstream.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/graph/graphviz.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/variant/static_visitor.hpp>
#include <graph_model/graph/tags.hpp>
#include <graph_model/graph/vertex_category_property_map.hpp>
#include <graph_model/graph/vertex_gid_property_map.hpp>
#include <graph_model/graph/vertex_identifier_property_map.hpp>
#include <observer_ptr.hpp>

#include <use_default.hpp>

namespace graph_model
{

  namespace graph
  {

    namespace detail
    {

      template < typename Tag >
      class graphviz_writer
      {
      
	struct attribute_base
	  : boost::static_visitor<std::string>
	{
	  attribute_base (std::string attribute) : attribute_(attribute) {}
	
	protected:
	  std::string attribute (std::string attribute) const
	  { return attribute_ + "=" + attribute + " "; }

	  std::string quoted_attribute (std::string attribute) const
	  { return this->attribute("\"" + attribute + "\""); }

	private:
	  std::string attribute_;
	};
	
	template < typename Graph >
	class vertex_writer
	{
	  using pointer = observer_ptr<Graph const>;
	  // using domain_type = traits::domain_t<Graph>;
	  // using vertex_properties = ::factor_graph::vertex_properties<domain_type>;

	  struct visitor_base : attribute_base
	  {
	    visitor_base (std::string attribute) : attribute_base(attribute) {}
	  };

	  template < typename VertexDescriptor >
	  struct label_visitor : visitor_base
	  {
	    label_visitor (Graph const& g, VertexDescriptor u)
	      : visitor_base("label"), p_(make_observer(g)), u_(u) {}

	    std::string operator () () const
	    // { return this->quoted_attribute(gid() + ":" + vertex_label()); }
	    { return this->quoted_attribute(vertex_label()); }

	  private:
	    // auto gid () const { return boost::lexical_cast<std::string>(graph().gid(u_)); }
	    // auto vertex_label () const { return get(&vertex_properties::label_,graph(),u_); }

	    auto vertex_label () const
	    { return vertex_identifier() + "\n" + vertex_category(); }

	    auto vertex_identifier () const // { return u_.identifier(); }
	    { return get(tag::vertex_identifier(),graph(),u_); }

	    auto vertex_category () const
	    { return boost::lexical_cast<std::string>(get(tag::vertex_category(),graph(),u_)); }

	    auto& graph () { return *p_; }
	    auto& graph () const { return *p_; }

	  private:
	    pointer p_;
	    VertexDescriptor u_;
	  };

	  template < typename VertexDescriptor >
	  static auto make_label_visitor (Graph const& g, VertexDescriptor u)
	  { return label_visitor<VertexDescriptor>(g,u); }

	  template < typename VertexDescriptor >
	  struct shape_visitor : visitor_base
	  {
	    shape_visitor (Graph const& g, VertexDescriptor u)
	      : visitor_base("shape"), p_(make_observer(g)), u_(u) {}

	    std::string operator () () const { return this->quoted_attribute(vertex_shape()); }

	  private:
	    struct visitor
	    {
	      template < typename T >
	      auto operator () (T t) const
	      {
		using is_factor = typename std::is_base_of<vertex::tag::factor,T>::type;
		return shape(t,is_factor());
	      }

	    private:
	      template < typename T >
	      auto shape (T t, std::false_type) const { return ""; }
	      template < typename T >
	      auto shape (T t, std::true_type) const { return "rectangle"; }
	    };

	  private:
	    auto vertex_shape () const
	    {
	      auto vertex_category = get(tag::vertex_category(),graph(),u_);
	      return boost::apply_visitor(visitor(),vertex_category);
	    }

	  private:
	    auto& graph () { return *p_; }
	    auto& graph () const { return *p_; }

	  private:
	    pointer p_;
	    VertexDescriptor u_;
	  };
	
	  template < typename VertexDescriptor >
	  static auto make_shape_visitor (Graph const& g, VertexDescriptor u)
	  { return shape_visitor<VertexDescriptor>(g,u); }
	
	  template < typename VertexDescriptor >
	  struct style_visitor : visitor_base
	  {
	    style_visitor (Graph const& g, VertexDescriptor u)
	      : visitor_base("style"), p_(make_observer(g)), u_(u) {}

	    template < typename VertexType >
	    std::string operator () (VertexType) const { return ""; }
	    std::string operator () (typename visitor_base::random_variable_vertex) const
	    { return vertex_style(); }

	  private:
	    auto vertex_style () const
	    {
	      // auto is_observed = get(&vertex_properties::is_observed_,graph(),u_);
	      auto is_observed = false;
	      return is_observed ? this->quoted_attribute("filled") : "";
	    }

	    auto& graph () { return *p_; }
	    auto& graph () const { return *p_; }

	  private:
	    pointer p_;
	    VertexDescriptor u_;
	  };
	
	  template < typename VertexDescriptor >
	  static auto make_style_visitor (Graph const& g, VertexDescriptor u)
	  { return style_visitor<VertexDescriptor>(g,u); }
	
	public:
	  vertex_writer () = default;
	  vertex_writer (Graph const& g) : p_(make_observer(g)) {}
	
	  template < typename VertexDescriptor >
	  void operator () (std::ostream& os, VertexDescriptor u) const
	  {
	    os << "[ ";
	    os << make_label_visitor(graph(),u)();
	    os << make_shape_visitor(graph(),u)();
	    os << "]";
	  }

	private:
	  auto& graph () { return *p_; }
	  auto& graph () const { return *p_; }
	
	private:
	  pointer p_;
	};

	template < typename Graph >
	static auto make_vertex_writer (Graph const& g) { return vertex_writer<Graph>(g); }

	template < typename Graph >
	class edge_writer
	{
	  using pointer = observer_ptr<Graph const>;
	  // using domain_type = traits::domain_t<Graph>;
	  // using edge_properties = ::factor_graph::edge_properties<domain_type>;

	  struct visitor_base : attribute_base
	  {
	    visitor_base (std::string attribute) : attribute_base(attribute) {}
	  };
	
	  struct color_visitor : visitor_base
	  {
	    color_visitor () : visitor_base("color") {}
	    std::string operator () (typename visitor_base::factor_edge) const
	    { return this->attribute("black"); }
	    template < typename EdgeType >
	    std::string operator () (EdgeType) const
	    { return this->attribute("gray"); }
	  };
	
	  class constraint_visitor : public visitor_base
	  {
	    using self = constraint_visitor;
	  
	    template < typename GraphTag, typename Visitor >
	    struct helper
	    {
	      static std::string apply (typename visitor_base::factor_edge, Visitor visitor)
	      { return visitor.attribute("false"); }
	      static std::string apply (typename visitor_base::variate_edge, self visitor)
	      { return visitor.attribute("false"); }
	      template < typename EdgeType >
	      static std::string apply (EdgeType, self) { return ""; }
	    };

	    template < typename Visitor >
	    struct helper< tag::graph_model, Visitor >
	    {
	      template < typename EdgeType >
	      static std::string apply (EdgeType, Visitor) { return ""; }
	    };

	  public:
	    constraint_visitor () : visitor_base("constraint") {}
	    template < typename EdgeType >
	    std::string operator () (EdgeType type) const
	    { return helper<Tag,self>::apply(type,*this); }
	  };
	
	  struct dir_visitor : visitor_base
	  {
	    dir_visitor () : visitor_base("dir") {}
	    template < typename EdgeType >
	    std::string operator () (EdgeType type) const { return apply(Tag(),type); }
	  private:
	    template < typename EdgeType >
	    std::string apply (tag::factor_graph, EdgeType) const
	    { return this->quoted_attribute("none"); }
	    template < typename GraphTag, typename EdgeType >
	    std::string apply (GraphTag, EdgeType) const { return ""; }
	  };
	
	  struct label_visitor : visitor_base
	  {
	    label_visitor () : visitor_base("label") {}
	    template < typename EdgeType >
	    std::string operator () (EdgeType) const { return ""; }
	  };
	
	  struct style_visitor : visitor_base
	  {
	    style_visitor () : visitor_base("style") {}
	    std::string operator () (typename visitor_base::expression_tree_edge) const
	    { return this->quoted_attribute("solid"); }
	    std::string operator () (typename visitor_base::factor_edge) const
	    { return this->quoted_attribute("solid"); }
	    std::string operator () (typename visitor_base::factor_graph_edge) const
	    { return this->quoted_attribute("dotted"); }
	    std::string operator () (typename visitor_base::proposer_edge) const
	    { return this->quoted_attribute("dotted"); }
	    std::string operator () (typename visitor_base::variate_edge) const
	    { return this->quoted_attribute("dotted"); }
	  };
	
	  struct weight_visitor : visitor_base
	  {
	    weight_visitor () : visitor_base("weight") {}
	    std::string operator () (typename visitor_base::expression_tree_edge) const
	    { return this->attribute("1000"); }
	    template < typename EdgeType >
	    std::string operator () (EdgeType) const { return ""; }
	  };
	
	public:
	  edge_writer () = default;
	  edge_writer (Graph const& g) : p_(make_observer(g)) {}

	  template < typename EdgeDescriptor >
	  void operator () (std::ostream& os, EdgeDescriptor e) const
	  {
	    os << "[ ";
	    os << "]";
	  }

	private:
	  template < typename EdgeDescriptor >
	  auto edge_type (EdgeDescriptor e) const
	  // { return get(&edge_properties::type_,graph(),e); }
	  { return std::string(); }

	  template < typename EdgeDescriptor >
	  std::string edge_label (EdgeDescriptor e) const
	  // { return get(&edge_properties::label_,graph(),e); }
	  { return std::string(); }
	
	private:
	  auto& graph () { return *p_; }
	  auto& graph () const { return *p_; }
	
	private:
	  pointer p_;
	};

	template < typename Graph >
	static auto make_edge_writer (Graph const& g) { return edge_writer<Graph>(g); }

	template < typename Graph >
	class graph_writer
	{
	  using pointer = observer_ptr<Graph const>;

	public:
	  graph_writer () = default;
	  graph_writer (Graph const& g) : p_(make_observer(g)) {}

	  void operator () (std::ostream& os) const {}

	private:
	  auto& graph () { return *p_; }
	  auto& graph () const { return *p_; }
	
	private:
	  pointer p_;
	};

	template < typename Graph >
	static auto make_graph_writer (Graph const& g) { return graph_writer<Graph>(g); }

      public:
	template < typename Graph >
	static void apply (boost::filesystem::path path, Graph const& g)
	{
	  using ofstream = boost::filesystem::ofstream;
	  ofstream os (path);
	  apply(os,g);
	}

	template < typename Graph >
	static void apply (std::ostream& os, Graph const& g) { apply(Tag(),os,g); }

      private:
	template < typename Graph >
	static void apply (use_default<>, std::ostream& os, Graph const& g)
	{
	  using boost::write_graphviz;
	  write_graphviz(os,g,
			 make_vertex_writer(g),
			 make_edge_writer(g),
			 make_graph_writer(g),
			 get(tag::vertex_gid(),g));
	}

	template < typename Graph >
	static void apply (tag::factor_graph, std::ostream& os, Graph const& g)
	{
	  using boost::write_graphviz;
	  auto factor_graph = make_factor_graph(g);
	  write_graphviz(os,factor_graph,
			 make_vertex_writer(g),
			 make_edge_writer(g),
			 make_graph_writer(g),
			 make_vertex_gid_property_map(g));
	}

	template < typename Graph >
	static void apply (tag::graph_model, std::ostream& os, Graph const& g)
	{
	  using boost::write_graphviz;
	  auto graph_model = make_graph_model(g);
	  write_graphviz(os,graph_model,
			 make_vertex_writer(g),
			 make_edge_writer(g),
			 make_graph_writer(g),
			 make_vertex_gid_property_map(g));
	}
      };

    } // namespace detail

    template < typename Tag, typename Destination, typename Graph >
    void write_graphviz (Destination&& destination, Graph&& graph)
    {
      detail::graphviz_writer<Tag>::apply
	(std::forward<Destination>(destination),std::forward<Graph>(graph));
    }

    template < typename Destination, typename Graph >
    void write_graphviz (Destination&& destination, Graph&& graph)
    {
      detail::graphviz_writer<use_default<>>::apply
	(std::forward<Destination>(destination),std::forward<Graph>(graph));
    }

  } // namespace graph

} // namespace graph_model

#endif
