/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_GRAPH_STREAMABLE_BASE_HPP
#define GRAPH_MODEL_GRAPH_STREAMABLE_BASE_HPP 1

#include <iosfwd>
#include <boost/graph/graph_traits.hpp>
#include <boost/graph/properties.hpp>
#include <graph_model/graph/color_map.hpp>
#include <graph_model/graph/graph_base.hpp>
#include <graph_model/graph_tags.hpp>
#include <graph_model/traits/edge_descriptor.hpp>
#include <graph_model/traits/vertex_descriptor.hpp>

namespace graph_model
{

  namespace graph
  {

    template < typename Derived, typename Tag, typename Base = graph_base >
    class streamable_base
      : public Base
    {
      using self = streamable_base;
      using vertex_descriptor = traits::vertex_descriptor_t<Tag>;
      using edge_descriptor = traits::edge_descriptor_t<Tag>;

    public:
      streamable_base () = default;
      template < typename Graph >
      streamable_base (Graph const& g, vertex_descriptor source)
	: Base(g,source)
      {}

      friend std::ostream& operator << (std::ostream& os, self const&) { return os; }

    private:
      Derived& derived () { return static_cast<Derived&>(*this); }
      Derived const& derived () const { return static_cast<Derived const&>(*this); }
    };

    template < typename Derived, typename Base >
    class streamable_base< Derived,::graph_model::tag::incidence_graph,Base >
      : public Base
    {
      using self = streamable_base;
      using vertex_descriptor = traits::vertex_descriptor_t<::graph_model::tag::incidence_graph>;

    public:
      streamable_base () = default;
      template < typename Graph >
      streamable_base (Graph const& g, vertex_descriptor source)
	: Base(g,source)
      {}

      friend std::ostream& operator << (std::ostream& os, Derived const& g)
      {
	os << "out_degree: " << out_degree(g.source(),g) << std::endl;
	auto pair = out_edges(g.source(),g);
	while (pair.first != pair.second)
	  {
	    std::cout
	      << source((*pair.first),g)
	      << " -> "
	      << target((*pair.first),g)
	      << std::endl;
	    ++pair.first;
	  }
	return os;
      }

    private:
      Derived& derived () { return static_cast<Derived&>(*this); }
      Derived const& derived () const { return static_cast<Derived const&>(*this); }
    };

    template < typename Derived, typename Base >
    class streamable_base< Derived,::graph_model::tag::vertex_list_graph,Base >
      : public Base
    {
      using self = streamable_base;
      using vertex_descriptor = traits::vertex_descriptor_t<::graph_model::tag::vertex_list_graph>;

    public:
      streamable_base () = default;
      template < typename Graph >
      streamable_base (Graph const& g, vertex_descriptor source)
	: Base(g,source)
      {}

      friend std::ostream& operator << (std::ostream& os, Derived const& g)
      {
	os << "vertices: " << num_vertices(g) << std::endl;
	auto pair = vertices(g);
	while (pair.first != pair.second)
	  {
	    std::cout << *pair.first << std::endl;
	    ++pair.first;
	  }
	return os;
      }

    private:
      Derived& derived () { return static_cast<Derived&>(*this); }
      Derived const& derived () const { return static_cast<Derived const&>(*this); }
    };

    template < typename Derived, typename Base >
    class streamable_base< Derived,::graph_model::tag::edge_list_graph,Base >
      : public Base
    {
      using self = streamable_base;
      using tag_type = ::graph_model::tag::edge_list_graph;
      using vertex_descriptor = traits::vertex_descriptor_t<tag_type>;
      using edge_descriptor = traits::edge_descriptor_t<tag_type>;

    public:
      streamable_base () = default;
      template < typename Graph >
      streamable_base (Graph const& g, vertex_descriptor source)
	: Base(g,source)
      {}

      friend std::ostream& operator << (std::ostream& os, Derived const& g)
      {
	std::cout << "edges: " << num_edges(g) << std::endl;
	auto pair = edges(g);
	while (pair.first != pair.second)
	  {
	    std::cout << *pair.first << std::endl;
	    ++pair.first;
	  }
	return os;
      }

    private:
      Derived& derived () { return static_cast<Derived&>(*this); }
      Derived const& derived () const { return static_cast<Derived const&>(*this); }
    };

    template < typename Derived, typename Base >
    class streamable_base< Derived,::graph_model::tag::vertex_and_edge_list_graph,Base >
      : public Base
    {
      using self = streamable_base;
      using tag_type = ::graph_model::tag::vertex_and_edge_list_graph;
      using vertex_descriptor = traits::vertex_descriptor_t<tag_type>;
      using edge_descriptor = traits::edge_descriptor_t<tag_type>;

    public:
      streamable_base () = default;
      template < typename Graph >
      streamable_base (Graph const& g, vertex_descriptor source)
	: Base(g,source)
      {}

      friend std::ostream& operator << (std::ostream& os, Derived const& g)
      {
	{
	  std::cout << "vertices: " << num_vertices(g) << std::endl;
	  auto pair = vertices(g);
	  while (pair.first != pair.second)
	    {
	      std::cout << *pair.first << std::endl;
	      ++pair.first;
	    }
	}
	{
	  std::cout << "edges: " << num_edges(g) << std::endl;
	  auto pair = edges(g);
	  while (pair.first != pair.second)
	    {
	      std::cout << *pair.first << std::endl;
	      ++pair.first;
	    }
	}
	return os;
      }

    private:
      Derived& derived () { return static_cast<Derived&>(*this); }
      Derived const& derived () const { return static_cast<Derived const&>(*this); }
    };

  } // namespace graph

} // namespace graph_model

#endif
