/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_GRAPH_VERTEX_PREDICATES_HPP
#define GRAPH_MODEL_GRAPH_VERTEX_PREDICATES_HPP 1

#include <type_traits>
#include <boost/variant/static_visitor.hpp>
#include <graph_model/graph/vertex_category_property_map.hpp>
#include <graph_model/vertex/tags.hpp>
#include <observer_ptr.hpp>

namespace graph_model
{

  namespace graph
  {

    struct always_true
    {
      template < typename VertexDescriptor >
      auto operator () (VertexDescriptor&&) const { return true; }
    };

    struct always_false
    {
      template < typename VertexDescriptor >
      auto operator () (VertexDescriptor&&) const { return false; }
    };

    template < typename Graph, typename Category >
    class is_vertex_category
    {
      struct visitor
	: boost::static_visitor<bool>
      {
	template < typename VertexCategory >
	bool operator () (VertexCategory) const
	{
	  using is_category
	    = typename std::is_base_of<Category,VertexCategory>::type;
	  return is_category::value;
	}
      };

      using pointer = observer_ptr<Graph const>;
 
    public:
      is_vertex_category () = default;
      is_vertex_category (Graph const& graph)
	: p_(make_observer(graph))
      {}

      template < typename VertexDescriptor >
      auto operator () (VertexDescriptor v) const
      {
	auto vertex_category = get(tag::vertex_category(),graph(),v);
	return boost::apply_visitor(visitor(),vertex_category);
      }

    private:
      auto& graph () { return *p_; }
      auto& graph () const { return *p_; }

    private:
      pointer p_;
    };
      
    template < typename Graph >
    class is_random_variable
      : public is_vertex_category<Graph,vertex::tag::random_variable>
    {
      using base = is_vertex_category<Graph,vertex::tag::random_variable>;

    public:
      is_random_variable () = default;
      is_random_variable (Graph const& graph)
	: base(graph)
      {}
    };
      
    template < typename Graph >
    class is_factor
      : public is_vertex_category<Graph,vertex::tag::factor>
    {
      using base = is_vertex_category<Graph,vertex::tag::factor>;

    public:
      is_factor () = default;
      is_factor (Graph const& graph)
	: base(graph)
      {}
    };

  } // namespace graph

} // namespace graph_model

#endif
